/*-
 * Copyright (c) 2004 Pawel Jakub Dawidek <pjd@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/bio.h>
#include <sys/sysctl.h>
#include <sys/malloc.h>
#include <geom/geom.h>
#include <geom/null/g_null.h>


SYSCTL_DECL(_kern_geom);
SYSCTL_NODE(_kern_geom, OID_AUTO, null, CTLFLAG_RW, 0, "GEOM_NULL stuff");
static u_int g_null_debug = 0;
SYSCTL_UINT(_kern_geom_null, OID_AUTO, debug, CTLFLAG_RW, &g_null_debug, 0,
    "Debug level");

static int g_null_destroy(struct g_geom *gp, boolean_t force);
static int g_null_destroy_geom(struct gctl_req *req, struct g_class *mp,
    struct g_geom *gp);
static void g_null_config(struct gctl_req *req, struct g_class *mp,
    const char *verb);

struct g_class g_null_class = {
	.name = G_NULL_CLASS_NAME,
	.ctlreq = g_null_config,
	.destroy_geom = g_null_destroy_geom
};


static void
g_null_orphan(struct g_consumer *cp)
{

	g_topology_assert();
	g_null_destroy(cp->geom, 1);
}

static void
g_null_start(struct bio *bp)
{
	struct g_geom *gp;
	struct bio *cbp;

	gp = bp->bio_to->geom;
	G_NULL_LOGREQ(bp, "Request received.");
	cbp = g_clone_bio(bp);
	if (cbp == NULL) {
		g_io_deliver(bp, ENOMEM);
		return;
	}
	cbp->bio_done = g_std_done;
	cbp->bio_offset = bp->bio_offset;
	cbp->bio_data = bp->bio_data;
	cbp->bio_length = bp->bio_length;
	cbp->bio_to = LIST_FIRST(&gp->provider);
	G_NULL_LOGREQ(cbp, "Sending request.");
	g_io_request(cbp, LIST_FIRST(&gp->consumer));
}

static int
g_null_access(struct g_provider *pp, int dr, int dw, int de)
{
	struct g_geom *gp;
	struct g_consumer *cp;
	int error;

	gp = pp->geom;
	cp = LIST_FIRST(&gp->consumer);
	error = g_access(cp, dr, dw, de);

	return (error);
}

static int
g_null_destroy(struct g_geom *gp, boolean_t force)
{
	struct g_provider *pp;

	g_topology_assert();
	pp = LIST_FIRST(&gp->provider);
	if (pp != NULL && (pp->acr != 0 || pp->acw != 0 || pp->ace != 0)) {
		if (force) {
			G_NULL_DEBUG(0, "Device %s is still open, so it "
			    "can't be definitely removed.", pp->name);
		} else {
			G_NULL_DEBUG(1, "Device %s is still open (r%dw%de%d).",
			    pp->name, pp->acr, pp->acw, pp->ace);
			return (EBUSY);
		}
	} else {
		G_NULL_DEBUG(0, "Device %s removed.", gp->name);
	}
	g_wither_geom(gp, ENXIO);

	return (0);
}

static int
g_null_destroy_geom(struct gctl_req *req, struct g_class *mp,
    struct g_geom *gp)
{

	return (g_null_destroy(gp, 0));
}

static void
g_null_ctl_create(struct gctl_req *req, struct g_class *mp)
{
	struct g_geom *gp;
	struct g_provider *newpp, *pp;
	struct g_consumer *cp;

	g_topology_assert();

	gp = NULL;
	newpp = NULL;
	cp = NULL;

	pp = gctl_get_provider(req, "provider");
	if (pp == NULL) {
		gctl_error(req, "Invalid provider.");
		return;
	}

	gp = g_new_geomf(mp, "%s%s", pp->name, G_NULL_SUFFIX);
	if (gp == NULL) {
		gctl_error(req, "Cannot create geom.");
		return;
	}
	gp->softc = NULL;
	gp->start = g_null_start;
	gp->spoiled = g_null_orphan;
	gp->orphan = g_null_orphan;
	gp->access = g_null_access;

	newpp = g_new_providerf(gp, gp->name);
	if (newpp == NULL) {
		gctl_error(req, "Cannot create provider.");
		goto fail;
	}
	newpp->mediasize = pp->mediasize;
	newpp->sectorsize = pp->sectorsize;

	cp = g_new_consumer(gp);
	if (cp == NULL) {
		gctl_error(req, "Cannot create consumer.");
		goto fail;
	}
	if (g_attach(cp, pp) != 0) {
		gctl_error(req, "Cannot attach to provider %s.", pp->name);
		goto fail;
	}

	g_error_provider(newpp, 0);
	G_NULL_DEBUG(0, "Device %s created.", gp->name);
	return;
fail:
	if (cp != NULL) {
		if (cp->provider != NULL)
			g_detach(cp);
		g_destroy_consumer(cp);
	}
	if (newpp != NULL)
		g_destroy_provider(pp);
	if (gp != NULL)
		g_destroy_geom(gp);
}

static void
g_null_ctl_destroy(struct gctl_req *req, struct g_geom *gp)
{
	int *force, error;

	g_topology_assert();
	force = gctl_get_paraml(req, "force", sizeof(*force));
	if (force == NULL) {
		gctl_error(req, "No 'force' argument");
		return;
	}
	error = g_null_destroy(gp, *force);
	if (error != 0) {
		gctl_error(req, "Cannot destroy device %s (error=%d).",
		    gp->name, error);
		return;
	}
}

static void
g_null_config(struct gctl_req *req, struct g_class *mp, const char *verb)
{
	struct g_geom *gp;

	g_topology_assert();

	if (strcmp(verb, "create device") == 0) {
		g_null_ctl_create(req, mp);
		return;
	}
	gp = gctl_get_geom(req, mp, "geom");
	if (gp == NULL) {
		gctl_error(req, "unknown device");
		return;
	}
	if (strcmp(verb, "destroy device") == 0) {
		g_null_ctl_destroy(req, gp);
		return;
	}

	gctl_error(req, "unknown verb");
}

DECLARE_GEOM_CLASS(g_null_class, g_null);
